############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries
import datetime

# external packages
import PyQt5.QtCore
import PyQt5.QtWidgets
from skyfield.toposlib import Topos

# local import
from base import transform


class Mount(object):
    """
    """

    def __init__(self):
        self.typeConnectionTexts = ['RS-232',
                                    'GPS/RS-232',
                                    'LAN',
                                    'WiFi',
                                    ]

        ms = self.app.mount.signals
        ms.locationDone.connect(self.updateLocGUI)
        ms.pointDone.connect(self.updatePointGUI)
        ms.pointDone.connect(self.updateTimeGUI)
        ms.settingDone.connect(self.updateSettingGUI)
        ms.settingDone.connect(self.updateSetStatGUI)
        ms.settingDone.connect(self.updateSetSyncGUI)
        ms.settingDone.connect(self.updateTrackingGui)

        self.ui.park.clicked.connect(self.changePark)
        self.ui.flipMount.clicked.connect(self.flipMount)
        self.ui.tracking.clicked.connect(self.changeTracking)
        self.ui.setLunarTracking.clicked.connect(self.setLunarTracking)
        self.ui.setSiderealTracking.clicked.connect(self.setSiderealTracking)
        self.ui.setSolarTracking.clicked.connect(self.setSolarTracking)
        self.ui.stop.clicked.connect(self.stop)
        self.clickable(self.ui.meridianLimitTrack).connect(self.setMeridianLimitTrack)
        self.clickable(self.ui.meridianLimitSlew).connect(self.setMeridianLimitSlew)
        self.clickable(self.ui.horizonLimitHigh).connect(self.setHorizonLimitHigh)
        self.clickable(self.ui.horizonLimitLow).connect(self.setHorizonLimitLow)
        self.clickable(self.ui.slewRate).connect(self.setSlewRate)
        self.clickable(self.ui.siteLatitude).connect(self.setLatitude)
        self.clickable(self.ui.siteLongitude).connect(self.setLongitude)
        self.clickable(self.ui.siteElevation).connect(self.setElevation)
        self.clickable(self.ui.statusUnattendedFlip).connect(self.setUnattendedFlip)
        self.clickable(self.ui.statusDualAxisTracking).connect(self.setDualAxisTracking)
        self.clickable(self.ui.statusWOL).connect(self.setWOL)
        self.clickable(self.ui.statusRefraction).connect(self.setRefraction)

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """
        config = self.app.config.get('mainW', {})
        self.ui.checkJ2000.setChecked(config.get('checkJ2000', False))
        self.ui.checkJNow.setChecked(config.get('checkJNow', False))
        self.updateLocGUI(self.app.mount.obsSite)
        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        config['checkJ2000'] = self.ui.checkJ2000.isChecked()
        config['checkJNow'] = self.ui.checkJNow.isChecked()
        return True

    def updatePointGUI(self, obs):
        """
        :param obs:
        :return:    True if ok for testing
        """
        if obs.Alt is not None:
            self.ui.ALT.setText('{0:5.2f}'.format(obs.Alt.degrees))

        else:
            self.ui.ALT.setText('-')

        if obs.Az is not None:
            self.ui.AZ.setText('{0:5.2f}'.format(obs.Az.degrees))

        else:
            self.ui.AZ.setText('-')

        isJ2000 = self.ui.checkJ2000.isChecked()
        isValid = obs.raJNow is not None
        isValid = isValid and obs.decJNow is not None
        isValid = isValid and obs.timeJD is not None
        if isJ2000 and isValid:
            ra, dec = transform.JNowToJ2000(obs.raJNow, obs.decJNow, obs.timeJD)

        else:
            ra = obs.raJNow
            dec = obs.decJNow

        if ra is not None:
            self.ui.RA.setText(self.formatHstrToText(ra))
            self.ui.RAfloat.setText(f'{ra.hours:3.4f}')

        else:
            self.ui.RA.setText('-')
            self.ui.RAfloat.setText('-')

        if dec is not None:
            self.ui.DEC.setText(self.formatDstrToText(dec))
            self.ui.DECfloat.setText(f'{dec.degrees:+3.4f}')

        else:
            self.ui.DEC.setText('-')
            self.ui.DECfloat.setText('-')

        if obs.pierside is not None:
            self.ui.pierside.setText('WEST' if obs.pierside == 'W' else 'EAST')

        else:
            self.ui.pierside.setText('-')

        if obs.haJNow is not None:
            self.ui.HA.setText(self.formatHstrToText(obs.haJNow))
            self.ui.HAfloat.setText(f'{obs.haJNow.hours:3.4f}')

        else:
            self.ui.HA.setText('-')
            self.ui.HAfloat.setText('-')

        return True

    def updateTimeGUI(self, obs):
        """
        :param obs:
        :return:    True if ok for testing
        """
        if obs.timeJD is not None:
            text = obs.timeJD.utc_strftime('%H:%M:%S')
            self.ui.timeUTC.setText('UTC: ' + text)

        if obs.timeSidereal is not None:
            siderealFormat = '{0:02.0f}:{1:02.0f}:{2:02.0f}'
            val = obs.timeSidereal.hms()
            siderealText = siderealFormat.format(*val)
            self.ui.timeSidereal.setText(siderealText)

        else:
            self.ui.timeSidereal.setText('-')

        return True

    def updateSetStatGUI(self, sett):
        """
        :param sett:
        :return:    True if ok for testing
        """
        if sett.UTCExpire is not None:
            ui = self.ui.UTCExpire
            ui.setText(sett.UTCExpire)

            # coloring if close to end:
            now = datetime.datetime.now()
            expire = datetime.datetime.strptime(sett.UTCExpire, '%Y-%m-%d')
            deltaYellow = datetime.timedelta(days=30)

            if now > expire:
                self.changeStyleDynamic(ui, 'color', 'red')

            elif now > expire - deltaYellow:
                self.changeStyleDynamic(ui, 'color', 'yellow')

            else:
                self.changeStyleDynamic(ui, 'color', '')

        self.guiSetText(self.ui.UTCExpire, 's', sett.UTCExpire)

        ui = self.ui.statusUnattendedFlip
        if sett.statusUnattendedFlip is None:
            text = '-'
            self.changeStyleDynamic(ui, 'status', '')

        elif sett.statusUnattendedFlip:
            text = 'ON'
            self.changeStyleDynamic(ui, 'status', 'on')

        else:
            text = 'OFF'
            self.changeStyleDynamic(ui, 'status', '')

        self.guiSetText(ui, 's', text)

        ui = self.ui.statusDualAxisTracking
        if sett.statusDualAxisTracking is None:
            text = '-'
            self.changeStyleDynamic(ui, 'status', '')

        elif sett.statusDualAxisTracking:
            text = 'ON'
            self.changeStyleDynamic(ui, 'status', 'on')

        else:
            text = 'OFF'
            self.changeStyleDynamic(ui, 'status', '')

        self.guiSetText(ui, 's', text)

        ui = self.ui.statusRefraction
        if sett.statusRefraction is None:
            text = '-'
            self.changeStyleDynamic(ui, 'status', '')
            self.changeStyleDynamic(self.ui.refractionTemp1, 'color', '')
            self.changeStyleDynamic(self.ui.refractionPress1, 'color', '')

        elif sett.statusRefraction:
            text = 'ON'
            self.changeStyleDynamic(ui, 'status', 'on')
            self.changeStyleDynamic(self.ui.refractionTemp1, 'color', '')
            self.changeStyleDynamic(self.ui.refractionPress1, 'color', '')

        else:
            text = 'OFF'
            self.changeStyleDynamic(ui, 'status', '')
            self.changeStyleDynamic(self.ui.refractionTemp1, 'color', 'yellow')
            self.changeStyleDynamic(self.ui.refractionPress1, 'color', 'yellow')

        self.guiSetText(ui, 's', text)

        return True

    def updateSetSyncGUI(self, sett):
        """
        :param sett:
        :return:    True if ok for testing
        """
        ui = self.ui.statusGPSSynced
        if sett.gpsSynced is None:
            text = '-'
            self.changeStyleDynamic(ui, 'status', '')
        elif sett.gpsSynced:
            text = 'YES'
            self.changeStyleDynamic(ui, 'status', 'on')
        else:
            text = 'NO'
            self.changeStyleDynamic(ui, 'status', '')
        self.guiSetText(ui, 's', text)

        ui = self.ui.statusWOL
        if sett.wakeOnLan is None:
            text = '-'
            self.changeStyleDynamic(ui, 'status', '')
        elif sett.wakeOnLan == 'On':
            text = 'ON'
            self.changeStyleDynamic(ui, 'status', 'on')
        else:
            text = 'OFF'
            self.changeStyleDynamic(ui, 'status', '')
        self.guiSetText(ui, 's', text)

        if sett.typeConnection is not None:
            text = self.typeConnectionTexts[sett.typeConnection]
            self.ui.mountTypeConnection.setText(text)

        return True

    def updateSettingGUI(self, sett):
        """
        :return:    True if ok for testing
        """
        self.guiSetText(self.ui.slewRate, '2.0f', sett.slewRate)
        self.guiSetText(self.ui.timeToFlip, '3.0f', sett.timeToFlip)
        self.guiSetText(self.ui.timeToMeridian, '3.0f', sett.timeToMeridian())
        self.guiSetText(self.ui.refractionTemp, '+4.1f', sett.refractionTemp)
        self.guiSetText(self.ui.refractionTemp1, '+4.1f', sett.refractionTemp)
        self.guiSetText(self.ui.refractionPress, '6.1f', sett.refractionPress)
        self.guiSetText(self.ui.refractionPress1, '6.1f', sett.refractionPress)
        self.guiSetText(self.ui.meridianLimitTrack, '3.0f', sett.meridianLimitTrack)
        self.guiSetText(self.ui.meridianLimitSlew, '3.0f', sett.meridianLimitSlew)
        self.guiSetText(self.ui.horizonLimitLow, '3.0f', sett.horizonLimitLow)
        self.guiSetText(self.ui.horizonLimitHigh, '3.0f', sett.horizonLimitHigh)

        return True

    def updateLocGUI(self, obs):
        """
        :param obs:
        :return:    True if ok for testing
        """
        if obs is None:
            return False

        location = obs.location
        if location is None:
            return False

        self.ui.siteLongitude.setText(self.formatLonToText(location.longitude))
        self.ui.siteLatitude.setText(self.formatLatToText(location.latitude))
        self.ui.siteElevation.setText(str(location.elevation.m))

        return True

    def updateTrackingGui(self, sett):
        """
        :param sett:
        :return:    True if ok for testing
        """
        if sett is None:
            return False
        if self.app.mount.obsSite.status is None:
            return False

        if self.app.mount.obsSite.status == 10:
            self.changeStyleDynamic(self.ui.followSat, 'running', 'true')
            self.changeStyleDynamic(self.ui.setLunarTracking, 'running', 'false')
            self.changeStyleDynamic(self.ui.setSiderealTracking, 'running', 'false')
            self.changeStyleDynamic(self.ui.setSolarTracking, 'running', 'false')

        elif sett.checkRateLunar():
            self.changeStyleDynamic(self.ui.followSat, 'running', 'false')
            self.changeStyleDynamic(self.ui.setLunarTracking, 'running', 'true')
            self.changeStyleDynamic(self.ui.setSiderealTracking, 'running', 'false')
            self.changeStyleDynamic(self.ui.setSolarTracking, 'running', 'false')

        elif sett.checkRateSidereal():
            self.changeStyleDynamic(self.ui.followSat, 'running', 'false')
            self.changeStyleDynamic(self.ui.setLunarTracking, 'running', 'false')
            self.changeStyleDynamic(self.ui.setSiderealTracking, 'running', 'true')
            self.changeStyleDynamic(self.ui.setSolarTracking, 'running', 'false')

        elif sett.checkRateSolar():
            self.changeStyleDynamic(self.ui.followSat, 'running', 'false')
            self.changeStyleDynamic(self.ui.setLunarTracking, 'running', 'false')
            self.changeStyleDynamic(self.ui.setSiderealTracking, 'running', 'false')
            self.changeStyleDynamic(self.ui.setSolarTracking, 'running', 'true')

        return True

    def changeTracking(self):
        """
        :return:
        """
        obs = self.app.mount.obsSite
        if obs.status == 0:
            suc = obs.stopTracking()
            if not suc:
                self.app.message.emit('Cannot stop tracking', 2)
            else:
                self.app.message.emit('Stopped tracking', 0)

        else:
            suc = obs.startTracking()
            if not suc:
                self.app.message.emit('Cannot start tracking', 2)
            else:
                self.app.message.emit('Started tracking', 0)

        return True

    def changePark(self):
        """
        :return:
        """
        obs = self.app.mount.obsSite
        if obs.status == 5:
            suc = obs.unpark()
            if not suc:
                self.app.message.emit('Cannot unpark mount', 2)
            else:
                self.app.message.emit('Mount unparked', 0)

        else:
            suc = obs.park()
            if not suc:
                self.app.message.emit('Cannot park mount', 2)
            else:
                self.app.message.emit('Mount parked', 0)

        return True

    def setLunarTracking(self):
        """
        :return:
        """
        sett = self.app.mount.setting
        suc = sett.setLunarTracking()
        if not suc:
            self.app.message.emit('Cannot set tracking to Lunar', 2)
            return False

        else:
            self.app.message.emit('Tracking set to Lunar', 0)
            return True

    def setSiderealTracking(self):
        """
        :return:
        """
        sett = self.app.mount.setting
        suc = sett.setSiderealTracking()
        if not suc:
            self.app.message.emit('Cannot set tracking to Sidereal', 2)
            return False

        else:
            self.app.message.emit('Tracking set to Sidereal', 0)
            return True

    def setSolarTracking(self):
        """
        :return:
        """
        sett = self.app.mount.setting
        suc = sett.setSolarTracking()
        if not suc:
            self.app.message.emit('Cannot set tracking to Solar', 2)
            return False

        else:
            self.app.message.emit('Tracking set to Solar', 0)
            return True

    def flipMount(self):
        """
        :return:
        """
        obs = self.app.mount.obsSite
        suc = obs.flip()
        if not suc:
            self.app.message.emit('Cannot flip mount', 2)
            return False

        else:
            self.app.message.emit('Mount flipped', 0)
            return True

    def stop(self):
        """
        :return:
        """
        obs = self.app.mount.obsSite
        suc = obs.stop()
        if not suc:
            self.app.message.emit('Cannot stop mount', 2)
            return False

        else:
            self.app.message.emit('Mount stopped', 0)
            return True

    def setMeridianLimitTrack(self):
        """
        :return:    success as bool if value could be changed
        """
        msg = PyQt5.QtWidgets.QMessageBox
        if not self.deviceStat.get('mount', ''):
            msg.critical(self,
                         'Error Message',
                         'Value cannot be set when mount not connected !')
            return False

        sett = self.app.mount.setting
        actValue = sett.meridianLimitTrack

        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getInt(self,
                               'Set Meridian Limit Track',
                               'Value (1-30):',
                               actValue,
                               1,
                               30,
                               1,
                               )

        if not ok:
            return False

        if sett.setMeridianLimitTrack(value):
            self.app.message.emit(f'Meridian Limit Track: [{value}]', 0)
            return True

        else:
            self.app.message.emit('Meridian Limit Track cannot be set', 2)
            return False

    def setMeridianLimitSlew(self):
        """
        :return:    success as bool if value could be changed
        """
        msg = PyQt5.QtWidgets.QMessageBox
        if not self.deviceStat.get('mount', ''):
            msg.critical(self,
                         'Error Message',
                         'Value cannot be set when mount not connected !')
            return False

        sett = self.app.mount.setting
        actValue = sett.meridianLimitSlew

        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getInt(self,
                               'Set Meridian Limit Slew',
                               'Value (0-30):',
                               actValue,
                               0,
                               30,
                               1,
                               )

        if not ok:
            return False

        if sett.setMeridianLimitSlew(value):
            self.app.message.emit(f'Meridian Limit Slew: [{value}]', 0)
            return True

        else:
            self.app.message.emit('Meridian Limit Slew cannot be set', 2)
            return False

    def setHorizonLimitHigh(self):
        """
        :return:    success as bool if value could be changed
        """
        msg = PyQt5.QtWidgets.QMessageBox
        if not self.deviceStat.get('mount', ''):
            msg.critical(self,
                         'Error Message',
                         'Value cannot be set when mount not connected !')
            return False

        sett = self.app.mount.setting
        actValue = sett.horizonLimitHigh

        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getInt(self,
                               'Set Horizon Limit High',
                               'Value (0-90):',
                               actValue,
                               0,
                               90,
                               1,
                               )

        if not ok:
            return False

        if sett.setHorizonLimitHigh(value):
            self.app.message.emit(f'Horizon Limit High: [{value}]', 0)
            return True

        else:
            self.app.message.emit('Horizon Limit High cannot be set', 2)
            return False

    def setHorizonLimitLow(self):
        """
        :return:    success as bool if value could be changed
        """
        msg = PyQt5.QtWidgets.QMessageBox
        if not self.deviceStat.get('mount', ''):
            msg.critical(self,
                         'Error Message',
                         'Value cannot be set when mount not connected !')
            return False

        sett = self.app.mount.setting
        actValue = sett.horizonLimitLow

        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getInt(self,
                               'Set Horizon Limit Low',
                               'Value (-5 - 90):',
                               actValue,
                               -5,
                               90,
                               1,
                               )

        if not ok:
            return False

        if sett.setHorizonLimitLow(value):
            self.app.message.emit(f'Horizon Limit Low: [{value}]', 0)
            return True

        else:
            self.app.message.emit('Horizon Limit Low cannot be set', 2)
            return False

    def setSlewRate(self):
        """
        :return:    success as bool if value could be changed
        """
        msg = PyQt5.QtWidgets.QMessageBox
        if not self.deviceStat.get('mount', ''):
            msg.critical(self,
                         'Error Message',
                         'Value cannot be set when mount not connected !')
            return False

        sett = self.app.mount.setting
        actValue = sett.slewRate

        minRate = sett.slewRateMin
        maxRate = sett.slewRateMax
        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getInt(self,
                               'Set Slew Rate',
                               f'Value ({minRate}-{maxRate}):',
                               actValue,
                               minRate,
                               maxRate,
                               1,
                               )

        if not ok:
            return False

        if sett.setSlewRate(value):
            self.app.message.emit(f'Slew Rate: [{value}]', 0)
            return True

        else:
            self.app.message.emit('Slew Rate cannot be set', 2)
            return False

    def setLongitude(self):
        """
        :return:    success as bool if value could be changed
        """
        obs = self.app.mount.obsSite
        if obs.location is None:
            return False

        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getText(self,
                                'Set Site Longitude',
                                'Format: <dd[EW] mm ss.s> or <[+-]d.d>, East is '
                                'positive',
                                PyQt5.QtWidgets.QLineEdit.Normal,
                                self.ui.siteLongitude.text(),
                                )
        if not ok:
            return False

        value = self.convertLonToAngle(value)

        if value is None:
            return False

        topo = Topos(longitude=value,
                     latitude=obs.location.latitude,
                     elevation_m=obs.location.elevation.m)
        obs.location = topo

        if not self.deviceStat.get('mount', ''):
            self.updateLocGUI(obs)
            return False

        if obs.setLongitude(value):
            self.app.message.emit(f'Longitude set to:    '
                                  f'[{self.ui.siteLongitude.text()}]', 0)
            self.app.mount.getLocation()
            return True

        else:
            self.app.message.emit('Longitude cannot be set', 2)
            return False

    def setLatitude(self):
        """
        :return:    success as bool if value could be changed
        """
        obs = self.app.mount.obsSite
        if obs.location is None:
            return False

        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getText(self,
                                'Set Site Latitude',
                                'Format: <dd[SN] mm ss.s> or <[+-]d.d>',
                                PyQt5.QtWidgets.QLineEdit.Normal,
                                self.ui.siteLatitude.text(),
                                )
        if not ok:
            return False

        value = self.convertLatToAngle(value)
        if value is None:
            return False

        topo = Topos(longitude=obs.location.longitude,
                     latitude=value,
                     elevation_m=obs.location.elevation.m)
        obs.location = topo

        if not self.deviceStat.get('mount', ''):
            self.updateLocGUI(obs)
            return False

        if obs.setLatitude(value):
            self.app.message.emit(f'Latitude set to:     '
                                  f'[{self.ui.siteLatitude.text()}]', 0)
            self.app.mount.getLocation()
            return True

        else:
            self.app.message.emit('Latitude cannot be set', 2)
            return False

    def setElevation(self):
        """
        :return:    success as bool if value could be changed
        """
        obs = self.app.mount.obsSite

        if obs.location is None:
            return False

        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getDouble(self,
                                  'Set Site Elevation',
                                  'Format: d.d',
                                  obs.location.elevation.m,
                                  0,
                                  8000,
                                  1,
                                  )
        if not ok:
            return False

        topo = Topos(longitude=obs.location.longitude,
                     latitude=obs.location.latitude,
                     elevation_m=value)
        obs.location = topo

        if not self.deviceStat.get('mount', ''):
            self.updateLocGUI(obs)
            return False

        if obs.setElevation(value):
            self.app.message.emit(f'Elevation set to:    [{value}]', 0)
            self.app.mount.getLocation()
            return True

        else:
            self.app.message.emit('Elevation cannot be set', 2)

    def setUnattendedFlip(self):
        """
        :return:    success as bool if value could be changed
        """
        msg = PyQt5.QtWidgets.QMessageBox
        if not self.deviceStat.get('mount', ''):
            msg.critical(self,
                         'Error Message',
                         'Value cannot be set when mount not connected !')
            return False

        sett = self.app.mount.setting

        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getItem(self,
                                'Set Unattended Flip',
                                'Value: On / Off',
                                ['ON', 'OFF'],
                                0,
                                False,
                                )
        if not ok:
            return False

        suc = sett.setUnattendedFlip(value == 'ON')
        if suc:
            self.app.message.emit(f'Unattended flip set to [{value}]', 0)

        else:
            self.app.message.emit('Unattended flip cannot be set', 2)
        return suc

    def setDualAxisTracking(self):
        """
        :return:    success as bool if value could be changed
        """
        msg = PyQt5.QtWidgets.QMessageBox
        if not self.deviceStat.get('mount', ''):
            msg.critical(self,
                         'Error Message',
                         'Value cannot be set when mount not connected !')
            return False

        sett = self.app.mount.setting

        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getItem(self,
                                'Set Dual Axis Tracking',
                                'Value: On / Off',
                                ['ON', 'OFF'],
                                0,
                                False,
                                )
        if not ok:
            return False

        suc = sett.setDualAxisTracking(value == 'ON')
        if suc:
            self.app.message.emit(f'Dual axis tracking set to [{value}]', 0)

        else:
            self.app.message.emit('Dual axis tracking cannot be set', 2)
        return suc

    def setWOL(self):
        """
        :return:    success as bool if value could be changed
        """
        msg = PyQt5.QtWidgets.QMessageBox
        if not self.deviceStat.get('mount', ''):
            msg.critical(self,
                         'Error Message',
                         'Value cannot be set when mount not connected !')
            return False

        sett = self.app.mount.setting

        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getItem(self,
                                'Set Wake On Lan',
                                'Value: On / Off',
                                ['ON', 'OFF'],
                                0,
                                False,
                                )
        if not ok:
            return False

        suc = sett.setWOL(value == 'ON')
        if suc:
            self.app.message.emit(f'WOL set to [{value}]', 0)

        else:
            self.app.message.emit('WOL cannot be set', 2)
        return suc

    def setRefraction(self):
        """
        :return:    success as bool if value could be changed
        """
        msg = PyQt5.QtWidgets.QMessageBox
        if not self.deviceStat.get('mount', ''):
            msg.critical(self,
                         'Error Message',
                         'Value cannot be set when mount not connected !')
            return False

        sett = self.app.mount.setting

        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getItem(self,
                                'Set Refraction Correction',
                                'Value: On / Off',
                                ['ON', 'OFF'],
                                0,
                                False,
                                )
        if not ok:
            return False

        suc = sett.setRefraction(value == 'ON')
        if suc:
            self.app.message.emit(f'Refraction correction set to [{value}]', 0)

        else:
            self.app.message.emit('Refraction correction cannot be set', 2)
        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries
import os
import time
import gzip
import shutil

# external packages
from PyQt5.QtCore import Qt, pyqtSignal
import requests

# local import
from gui.utilities import toolsQtWidget
from base.tpool import Worker
from gui.widgets.downloadPopup_ui import Ui_DownloadPopup


class DownloadPopup(toolsQtWidget.MWidget):
    """
    the DevicePopup window class handles

    """

    __all__ = ['DownloadPopup',
               ]

    signalProgress = pyqtSignal(object)
    signalProgressBarColor = pyqtSignal(object)

    def __init__(self,
                 parentWidget,
                 url='',
                 dest='',
                 callBack=None,
                 ):

        super().__init__()
        self.ui = Ui_DownloadPopup()
        self.ui.setupUi(self)
        self.initUI()
        self.callBack = callBack
        self.worker = None
        self.setWindowModality(Qt.ApplicationModal)
        x = parentWidget.x() + int((parentWidget.width() - self.width()) / 2)
        y = parentWidget.y() + int((parentWidget.height() - self.height()) / 2)
        self.move(x, y)

        baseName = os.path.basename(url)
        self.setWindowTitle(f'Downloading [{baseName}]')

        self.threadPool = parentWidget.threadPool
        self.signalProgress.connect(self.setProgressBarToValue)
        self.signalProgressBarColor.connect(self.setProgressBarColor)
        self.downloadFile(url, dest)
        self.show()

    def setProgressBarColor(self, color):
        """
        :param color:
        :return:
        """
        css = 'QProgressBar::chunk {background-color: ' + color + ';}'
        self.ui.progressBar.setStyleSheet(css)
        return True

    def setProgressBarToValue(self, progressPercent):
        """
        :param progressPercent:
        :return: True for test purpose
        """
        self.ui.progressBar.setValue(progressPercent)
        return True

    def getFileFromUrl(self, url, dest):
        """
        :param url:
        :param dest:
        :return:
        """
        r = requests.get(url, stream=True, timeout=1)
        totalSizeBytes = int(r.headers.get('content-length', 1))

        with open(dest, 'wb') as f:
            for n, chunk in enumerate(r.iter_content(512)):
                progressPercent = int(n * 512 / totalSizeBytes * 100)
                self.signalProgress.emit(progressPercent)

                if chunk:
                    f.write(chunk)

            self.signalProgress.emit(100)
        return True

    @staticmethod
    def unzipFile(dest):
        """
        :param dest:
        :return: True for test purpose
        """
        destUnzip = dest[:-3]
        with gzip.open(dest, 'rb') as f_in:
            with open(destUnzip, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)

        if os.path.isfile(dest):
            os.remove(dest)

        return True

    def downloadFileWorker(self, url, dest):
        """
        :param url:
        :param dest:
        :return:
        """
        if not os.path.dirname(dest):
            return False

        if os.path.isfile(dest):
            os.remove(dest)

        try:
            self.getFileFromUrl(url, dest)

        except TimeoutError:
            self.log.debug(f'Download [{url}] timed out')
            self.signalProgressBarColor.emit('red')
            time.sleep(1)
            return False

        except Exception as e:
            self.log.warning(f'Download [{url}] error: {e}')
            self.signalProgressBarColor.emit('red')
            time.sleep(1)
            return False

        else:
            self.signalProgressBarColor.emit('green')

        finally:
            time.sleep(1)

        try:
            self.unzipFile(dest)

        except Exception as e:
            self.log.warning(f'Error in unzip [{url}], {e}')
            return False
        return True

    def processResult(self, result):
        """
        :return:
        """
        if result:
            self.callBack()

        self.close()
        return True

    def downloadFile(self, url, dest):
        """
        :param url:
        :param dest:
        :return:
        """
        self.worker = Worker(self.downloadFileWorker, url=url, dest=dest)

        if self.callBack:
            self.worker.signals.result.connect(self.processResult)

        else:
            self.worker.signals.result.connect(self.close)

        self.threadPool.start(self.worker)

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages
import PyQt5
from PyQt5.QtCore import QRect
import numpy as np
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt

# local import
from gui.utilities import toolsQtWidget
from gui.widgets import hemisphere_ui
from gui.extWindows.hemisphereWext import HemisphereWindowExt


class HemisphereWindow(toolsQtWidget.MWidget, HemisphereWindowExt):
    """
    the hemisphere window class handles all interaction with model points show / edit etc.
    the z orders is aligned as follows:
    there are two planes static / moving where the moving is behind the static one. the
    static on has to be transparent.

    on the static plane we have (and set to the z order)
        - horizon               0
        - horizon limits        0
        - meridian limit track  10
        - meridian limit slew   15
        - celestial path        20
        - alignment stars       30
        - build points          40
        - checked build points  50

    on the moving plane we have (and set to the z order)
        - dome                  0
        - pointing marker       10

    """

    __all__ = ['HemisphereWindow',
               ]

    def __init__(self, app):
        super().__init__()

        self.app = app
        self.ui = hemisphere_ui.Ui_HemisphereDialog()
        self.ui.setupUi(self)
        self.initUI()
        self.mutexDraw = PyQt5.QtCore.QMutex()
        self.operationMode = 'normal'
        self.statusDAT = None

        self.MODE = dict(
            normal=dict(horMarker='None',
                        horColor=self.M_GREEN,
                        buildPColor=self.M_GREEN_H,
                        starSize=6,
                        starColor=self.M_YELLOW_L,
                        starAnnColor=self.M_WHITE_L),

            build=dict(horMarker='None',
                       horColor=self.M_GREEN,
                       buildPColor=self.M_PINK_H,
                       starSize=6,
                       starColor=self.M_YELLOW_L,
                       starAnnColor=self.M_WHITE_L),

            horizon=dict(horMarker='o',
                         horColor=self.M_PINK_H,
                         buildPColor=self.M_GREEN_L,
                         starSize=6,
                         starColor=self.M_YELLOW_L,
                         starAnnColor=self.M_WHITE_L),

            star=dict(horMarker='None',
                      horColor=self.M_GREEN_LL,
                      buildPColor=self.M_GREEN_L,
                      starSize=12,
                      starColor=self.M_YELLOW_H,
                      starAnnColor=self.M_WHITE_H)
        )

        # attributes to be stored in class
        self.pointerAltAz = None
        self.pointerDome = None
        self.pointsBuild = None
        self.starsAlign = None
        self.starsAlignAnnotate = None
        self.meridianSlew = None
        self.meridianTrack = None
        self.horizonLimitHigh = None
        self.horizonLimitLow = None
        self.celestialPath = None
        self.celestialPolarPath = None
        self.meridianSlewParam = None
        self.meridianTrackParam = None
        self.horizonLimitHighParam = None
        self.horizonLimitLowParam = None
        self.pointerPolarAltAz = None
        self.pointsBuildAnnotate = None
        self.pointsPolarBuild = None
        self.pointsPolarBuildAnnotate = None
        self.closingWindow = False

        self.ui.hemisphereMove.stackUnder(self.ui.hemisphere)
        self.hemisphereMat = self.embedMatplot(self.ui.hemisphere)
        self.hemisphereMatMove = self.embedMatplot(self.ui.hemisphereMove)

        self.ui.polar.stackUnder(self.ui.polarMove)
        self.polarMat = self.embedMatplot(self.ui.polar)
        self.polarMatMove = self.embedMatplot(self.ui.polarMove)

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """

        if 'hemisphereW' not in self.app.config:
            self.app.config['hemisphereW'] = {}

        config = self.app.config['hemisphereW']
        x = config.get('winPosX', 60)
        y = config.get('winPosY', 60)

        if x > self.screenSizeX:
            x = 0

        if y > self.screenSizeY:
            y = 0

        self.move(x, y)
        height = config.get('height', 600)
        width = config.get('width', 800)
        self.resize(width, height)
        self.ui.checkShowSlewPath.setChecked(config.get('checkShowSlewPath', False))
        self.ui.checkShowMeridian.setChecked(config.get('checkShowMeridian', False))
        self.ui.checkShowCelestial.setChecked(config.get('checkShowCelestial', False))
        self.ui.checkShowAlignStar.setChecked(config.get('checkShowAlignStar', False))
        self.ui.checkUseHorizon.setChecked(config.get('checkUseHorizon', True))
        self.ui.showPolar.setChecked(config.get('showPolar', False))

        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        if 'hemisphereW' not in self.app.config:
            self.app.config['hemisphereW'] = {}

        config = self.app.config['hemisphereW']
        config['winPosX'] = self.pos().x()
        config['winPosY'] = self.pos().y()
        config['height'] = self.height()
        config['width'] = self.width()
        config['checkShowSlewPath'] = self.ui.checkShowSlewPath.isChecked()
        config['checkShowMeridian'] = self.ui.checkShowMeridian.isChecked()
        config['checkShowCelestial'] = self.ui.checkShowCelestial.isChecked()
        config['checkShowAlignStar'] = self.ui.checkShowAlignStar.isChecked()
        config['checkUseHorizon'] = self.ui.checkUseHorizon.isChecked()
        config['showPolar'] = self.ui.showPolar.isChecked()

        return True

    def closeEvent(self, closeEvent):
        """
        :param closeEvent:
        :return:
        """
        self.closingWindow = True
        self.app.update10s.disconnect(self.updateAlignStar)
        self.app.update1s.disconnect(self.hemisphereMatMove.figure.canvas.draw)
        self.app.redrawHemisphere.disconnect(self.drawHemisphere)
        self.app.updatePointMarker.disconnect(self.updatePointMarker)
        self.app.updatePointMarker.disconnect(self.updatePolarPointMarker)
        self.app.mount.signals.settingDone.disconnect(self.updateOnChangedParams)
        self.storeConfig()

        self.ui.checkEditNone.setChecked(True)
        self.setOperationMode()

        self.ui.checkShowSlewPath.clicked.disconnect(self.drawHemisphere)
        self.ui.checkShowMeridian.clicked.disconnect(self.drawHemisphere)
        self.ui.checkShowCelestial.clicked.disconnect(self.drawHemisphere)
        self.ui.checkUseHorizon.clicked.disconnect(self.drawHemisphere)
        self.ui.checkEditNone.clicked.disconnect(self.setOperationMode)
        self.ui.checkEditHorizonMask.clicked.disconnect(self.setOperationMode)
        self.ui.checkEditBuildPoints.clicked.disconnect(self.setOperationMode)
        self.ui.checkPolarAlignment.clicked.disconnect(self.setOperationMode)
        self.ui.checkShowAlignStar.clicked.disconnect(self.drawHemisphere)
        self.ui.showPolar.clicked.disconnect(self.togglePolar)
        self.app.mount.signals.pointDone.disconnect(self.updatePointerAltAz)
        self.app.mount.signals.pointDone.disconnect(self.updatePointerPolarAltAz)
        self.app.dome.signals.azimuth.disconnect(self.updateDome)
        self.app.dome.signals.deviceDisconnected.disconnect(self.updateDome)
        self.app.dome.signals.serverDisconnected.disconnect(self.updateDome)

        plt.close(self.hemisphereMat.figure)
        plt.close(self.polarMat.figure)
        plt.close(self.hemisphereMatMove.figure)
        plt.close(self.polarMatMove.figure)

        super().closeEvent(closeEvent)

    def resizeEvent(self, event):
        """
        we are using the resize event to reset the timer, which means waiting for
        RESIZE_FINISHED_TIMEOUT in total before redrawing the complete hemisphere.
        as we are using a 0.1s cyclic timer.

        :param event:
        :return:
        """

        geometry = self.ui.hemisphere.geometry()
        newGeometry = QRect(0, 0, geometry.width(), geometry.height())
        self.ui.hemisphereMove.setGeometry(newGeometry)

        if self.ui.showPolar.isChecked():
            geometry = self.ui.polar.geometry()
            newGeometry = QRect(0, 0, geometry.width(), geometry.height())
            self.ui.polarMove.setGeometry(newGeometry)

        super().resizeEvent(event)

    def showWindow(self):
        """

        :return:
        """

        self.app.update10s.connect(self.updateAlignStar)
        self.app.update1s.connect(self.hemisphereMatMove.figure.canvas.draw)
        self.app.update1s.connect(self.polarMatMove.figure.canvas.draw)
        self.app.redrawHemisphere.connect(self.drawHemisphere)
        self.app.updatePointMarker.connect(self.updatePointMarker)
        self.app.updatePointMarker.connect(self.updatePolarPointMarker)
        self.app.mount.signals.settingDone.connect(self.updateOnChangedParams)
        self.app.mount.signals.pointDone.connect(self.updatePointerAltAz)
        self.app.mount.signals.pointDone.connect(self.updatePointerPolarAltAz)
        self.app.dome.signals.azimuth.connect(self.updateDome)
        self.app.dome.signals.deviceDisconnected.connect(self.updateDome)
        self.app.dome.signals.serverDisconnected.connect(self.updateDome)

        self.ui.checkShowSlewPath.clicked.connect(self.drawHemisphere)
        self.ui.checkUseHorizon.clicked.connect(self.drawHemisphere)
        self.ui.checkShowAlignStar.clicked.connect(self.drawHemisphere)
        self.ui.checkShowMeridian.clicked.connect(self.drawHemisphere)
        self.ui.checkShowCelestial.clicked.connect(self.drawHemisphere)
        self.ui.checkEditNone.clicked.connect(self.setOperationMode)
        self.ui.checkEditHorizonMask.clicked.connect(self.setOperationMode)
        self.ui.checkEditBuildPoints.clicked.connect(self.setOperationMode)
        self.ui.checkPolarAlignment.clicked.connect(self.setOperationMode)
        self.ui.showPolar.clicked.connect(self.togglePolar)
        self.ui.addPositionToHorizon.clicked.connect(self.addHorizonPointManual)

        self.hemisphereMat.figure.canvas.mpl_connect('button_press_event',
                                                     self.onMouseDispatcher)
        self.hemisphereMat.figure.canvas.mpl_connect('motion_notify_event',
                                                     self.showMouseCoordinates)
        self.togglePolar()
        self.show()

        return True

    def togglePolar(self):
        """

        :return: success
        """

        if self.ui.showPolar.isChecked():
            self.ui.polar.setMaximumSize(16777215, 16777215)
            self.ui.polar.setVisible(True)

        else:
            self.ui.polar.setFixedWidth(1)
            self.ui.polar.setVisible(False)

        self.drawHemisphere()

        return True

    def updateOnChangedParams(self, sett):
        """

        :param sett:
        :return: status redraw
        """

        needRedraw = False

        if self.meridianSlewParam != sett.meridianLimitSlew:
            self.meridianSlewParam = sett.meridianLimitSlew
            needRedraw = True

        if self.meridianTrackParam != sett.meridianLimitTrack:
            self.meridianTrackParam = sett.meridianLimitTrack
            needRedraw = True

        if self.horizonLimitHighParam != sett.horizonLimitHigh:
            self.horizonLimitHighParam = sett.horizonLimitHigh
            needRedraw = True

        if self.horizonLimitLowParam != sett.horizonLimitLow:
            self.horizonLimitLowParam = sett.horizonLimitLow
            needRedraw = True

        if needRedraw:
            self.drawHemisphere()

        return needRedraw

    def updatePointerAltAz(self):
        """
        :return: success
        """

        if self.pointerAltAz is None:
            return False

        obsSite = self.app.mount.obsSite

        if obsSite.Alt is None or obsSite.Az is None:
            self.pointerAltAz.set_visible(False)
            return False

        else:
            self.pointerAltAz.set_visible(True)

        alt = obsSite.Alt.degrees
        az = obsSite.Az.degrees
        self.pointerAltAz.set_data((az, alt))

        return True

    def updatePointerPolarAltAz(self):
        """
        :return: success
        """

        if self.pointerPolarAltAz is None:
            return False

        obsSite = self.app.mount.obsSite

        if obsSite.Alt is None or obsSite.Az is None:
            self.pointerPolarAltAz.set_visible(False)
            return False

        else:
            self.pointerPolarAltAz.set_visible(True)

        alt = 90 - obsSite.Alt.degrees
        az = obsSite.Az.radians
        self.pointerPolarAltAz.set_data((az, alt))

        return True

    def updateDome(self, azimuth):
        """
        updateDome is called whenever an update of coordinates from dome are given.
        it takes the actual values and corrects the point in window if window is in
        show status.

        :param azimuth:
        :return: success
        """

        if self.pointerDome is None:
            return False

        if not isinstance(azimuth, (int, float)):
            self.pointerDome.set_visible(False)
            return False

        visible = self.app.deviceStat['dome']

        self.pointerDome.set_xy((azimuth - 15, 1))
        self.pointerDome.set_visible(visible)

        return True

    def getMarkerStatusParams(self, active, index):
        """

        :param active:
        :param index:
        :return: values
        """

        if active:
            marker = self.markerPoint()
            color = self.MODE[self.operationMode]['buildPColor']
            markersize = 9
            annotationText = f'{index + 1:2d}'

        else:
            marker = '$\u2714$'
            color = self.M_YELLOW
            markersize = 11
            annotationText = f'{index + 1:2d}'

        return marker, markersize, color, annotationText

    def updatePointMarker(self):
        """

        :return: success
        """

        for index, point in enumerate(self.app.data.buildP):
            active = point[2]

            marker, markersize, color, _ = self.getMarkerStatusParams(active, index)

            self.pointsBuild[index].set_marker(marker)
            self.pointsBuild[index].set_markersize(markersize)
            self.pointsBuild[index].set_color(color)
            self.pointsBuildAnnotate[index].set_color(color)

        self.hemisphereMat.figure.canvas.draw()

        return True

    def updatePolarPointMarker(self):
        """

        :return: success
        """
        if not self.pointsPolarBuild:
            return False

        for index, point in enumerate(self.app.data.buildP):
            active = point[2]

            marker, markersize, color, _ = self.getMarkerStatusParams(active, index)

            self.pointsPolarBuild[index].set_marker(marker)
            self.pointsPolarBuild[index].set_markersize(markersize)
            self.pointsPolarBuild[index].set_color(color)
            self.pointsPolarBuildAnnotate[index].set_color(color)

        self.polarMat.figure.canvas.draw()

        return True

    def updateAlignStar(self):
        """
        updateAlignStar is called whenever an update of coordinates from mount are
        given. it takes the actual values and corrects the point in window if window is in
        show status.
        If the object is not created, the routing returns false.

        due to the fact that all annotation are only shown if in axes when coordinate
        are in data, after some time, no annotation will be shown, because just moved.
        therefore we add each time the annotation again.

        :return: success
        """

        if self.starsAlign is None:
            return False

        if self.starsAlignAnnotate is None:
            return False

        axes = self.hemisphereMat.figure.axes[0]
        hip = self.app.hipparcos
        hip.calculateAlignStarPositionsAltAz()
        self.starsAlign.set_data(hip.az, hip.alt)

        for i, starAnnotation in enumerate(self.starsAlignAnnotate):
            starAnnotation.remove()
        self.starsAlignAnnotate = list()

        for alt, az, name in zip(hip.alt, hip.az, hip.name):
            annotation = axes.annotate(name,
                                       xy=(az, alt),
                                       xytext=(2, 2),
                                       textcoords='offset points',
                                       xycoords='data',
                                       color=self.M_WHITE_L,
                                       fontsize=12,
                                       zorder=30,
                                       clip_on=True,
                                       )
            self.starsAlignAnnotate.append(annotation)

        self.hemisphereMat.figure.canvas.draw()

        return True

    def clearHemisphere(self):
        """
        clearHemisphere is called when after startup the location of the mount is changed
        to reconstruct correctly the hemisphere window

        :return: True for test Purpose
        """

        self.pointsBuild = None
        self.app.data.clearBuildP()
        self.drawHemisphere()

        return True

    def staticHorizon(self, axes=None, polar=False):
        """

        :param axes:
        :param polar:
        :return:
        """
        if not self.app.data.horizonP:
            return False

        alt, az = zip(*self.app.data.horizonP)

        if polar:
            az = np.radians(az)
            alt = np.array(alt)
            azF = np.radians(range(0, 361, 5))
            altF = np.interp(azF, az, alt)

            axes.fill(azF,
                      90 - altF,
                      color=self.M_TRANS,
                      alpha=0.5,
                      zorder=0)

            axes.plot(az,
                      90 - alt,
                      color=self.MODE['normal']['horColor'],
                      marker=self.MODE[self.operationMode]['horMarker'],
                      alpha=0.5,
                      zorder=0,
                      ls='none')

            axes.plot(azF,
                      90 - altF,
                      color=self.MODE['normal']['horColor'],
                      marker='',
                      alpha=0.5,
                      zorder=0,
                      lw=3)

        else:
            alt = np.array(alt)
            az = np.array(az)
            altF = np.concatenate([[0], [alt[0]], alt, [alt[-1]], [0]])
            azF = np.concatenate([[0], [0], az, [360], [360]])
            axes.fill(azF,
                      altF,
                      color=self.M_GREEN_LL,
                      alpha=0.5,
                      zorder=0)

            axes.plot(az,
                      alt,
                      color=self.MODE[self.operationMode]['horColor'],
                      marker=self.MODE[self.operationMode]['horMarker'],
                      alpha=0.5,
                      zorder=0,
                      ls='none')

            axes.plot(az,
                      alt,
                      color=self.MODE[self.operationMode]['horColor'],
                      marker='',
                      alpha=0.5,
                      zorder=0,
                      lw=3)

        return True

    def staticModelData(self, axes=None, polar=False):
        """

        :param axes:
        :param polar:
        :return:
        """
        points = self.app.data.buildP

        if not points:
            return False

        if polar:
            self.pointsPolarBuild = list()
            self.pointsPolarBuildAnnotate = list()

        else:
            self.pointsBuild = list()
            self.pointsBuildAnnotate = list()

        for index, point in enumerate(points):
            az = point[1]
            alt = point[0]
            active = point[2]

            marker, markersize, color, text = self.getMarkerStatusParams(active, index)

            if self.ui.checkShowSlewPath.isChecked() and index > 0:
                if polar:
                    axes.plot(np.radians((points[index - 1][1], points[index][1])),
                              (90 - points[index - 1][0], 90 - points[index][0]),
                              ls=':', lw=1, color=self.M_WHITE, zorder=40)

                else:
                    axes.plot((points[index - 1][1], points[index][1]),
                              (points[index - 1][0], points[index][0]),
                              ls=':', lw=1, color=self.M_WHITE, zorder=40)

            if polar:
                p, = axes.plot(np.radians(az), 90 - alt,
                               marker=marker,
                               markersize=markersize,
                               fillstyle='none',
                               color=self.MODE['normal']['buildPColor'],
                               zorder=40,
                               )
                self.pointsPolarBuild.append(p)

            else:
                p, = axes.plot(az, alt,
                               marker=marker,
                               markersize=markersize,
                               fillstyle='none',
                               color=self.MODE[self.operationMode]['buildPColor'],
                               zorder=40,
                               )
                self.pointsBuild.append(p)

            if polar:
                annotation = axes.annotate(text,
                                           xy=(np.radians(az), 90 - alt),
                                           xytext=(2, -10),
                                           textcoords='offset points',
                                           color=color,
                                           zorder=40,
                                           )
                self.pointsPolarBuildAnnotate.append(annotation)

            else:
                annotation = axes.annotate(text,
                                           xy=(az, alt),
                                           xytext=(2, -10),
                                           textcoords='offset points',
                                           color=color,
                                           zorder=40,
                                           )
                self.pointsBuildAnnotate.append(annotation)

        return True

    def staticCelestialEquator(self, axes=None, polar=False):
        """
        :param axes:
        :param polar:
        :return:
        """
        celestial = self.app.data.generateCelestialEquator()
        if not celestial:
            return False

        alt, az = zip(*celestial)
        alt = np.array(alt)

        if polar:
            self.celestialPath, = axes.plot(np.radians(az),
                                            90 - alt,
                                            '.',
                                            markersize=1,
                                            fillstyle='none',
                                            zorder=20,
                                            color=self.M_WHITE_L)
        else:
            self.celestialPath, = axes.plot(az,
                                            alt,
                                            '.',
                                            markersize=1,
                                            fillstyle='none',
                                            zorder=20,
                                            color=self.M_WHITE_L)
        return True

    def staticMeridianLimits(self, axes=None):
        """

        :param axes:
        :return:
        """

        slew = self.app.mount.setting.meridianLimitSlew
        track = self.app.mount.setting.meridianLimitTrack

        if slew is None or track is None:
            return False

        self.meridianSlew = mpatches.Rectangle((180 - slew, 0),
                                               2 * slew,
                                               90,
                                               zorder=15,
                                               color=self.M_YELLOW,
                                               alpha=0.5)
        axes.add_patch(self.meridianSlew)

        self.meridianTrack = mpatches.Rectangle((180 - track, 0),
                                                2 * track,
                                                90,
                                                zorder=10,
                                                color=self.M_YELLOW_L,
                                                alpha=0.5)
        axes.add_patch(self.meridianTrack)

        return True

    def staticHorizonLimits(self, axes=None, polar=False):
        """

        :param axes:
        :param polar:
        :return:
        """
        if self.app.mount.setting.horizonLimitHigh is not None:
            high = self.app.mount.setting.horizonLimitHigh

        else:
            high = 90

        if self.app.mount.setting.horizonLimitLow is not None:
            low = self.app.mount.setting.horizonLimitLow

        else:
            low = 0

        self.horizonLimitHigh = mpatches.Rectangle((0, high),
                                                   360,
                                                   90 - high,
                                                   zorder=0,
                                                   color=self.M_RED,
                                                   alpha=0.5,
                                                   visible=True)

        self.horizonLimitLow = mpatches.Rectangle((0, 0),
                                                  360,
                                                  low,
                                                  zorder=0,
                                                  color=self.M_RED,
                                                  alpha=0.5,
                                                  visible=True)
        axes.add_patch(self.horizonLimitHigh)
        axes.add_patch(self.horizonLimitLow)

        return True

    def drawHemisphereStatic(self, axes=None, polar=False):
        """
         drawHemisphereStatic renders the static part of the hemisphere window and puts
         all drawing on the static plane. the content consist of:
            - modeldata points
            - horizon mask
            - celestial paths
            - meridian limits
        with all their styles an coloring

        :param axes: matplotlib axes object
        :param polar: flag if polar should be drawn
        :return: success
        """
        if polar:
            if self.ui.checkShowCelestial.isChecked():
                self.staticCelestialEquator(axes=axes, polar=polar)

            if self.ui.checkUseHorizon.isChecked():
                self.staticHorizon(axes=axes, polar=polar)

            self.staticModelData(axes=axes, polar=polar)

        else:
            if self.ui.checkUseHorizon.isChecked():
                self.staticHorizon(axes=axes)
                self.staticHorizonLimits(axes=axes)

            if self.ui.checkShowCelestial.isChecked():
                self.staticCelestialEquator(axes=axes)

            if self.ui.checkShowMeridian.isChecked():
                self.staticMeridianLimits(axes=axes)

            self.staticModelData(axes=axes, polar=polar)

        return True

    def drawHemisphereMoving(self, axes=None, polar=False):
        """
        drawHemisphereMoving is rendering the moving part which consists of:
            - pointer: where the mount points to
            - dome widget: which shows the position of the dome opening
        the dynamic ones are located on a separate plane to improve rendering speed,
        because we update this part very often.

        :param axes: matplotlib axes object
        :param polar: flag if polar should be drawn
        :return:
        """

        if polar:
            self.pointerPolarAltAz, = axes.plot(np.radians(180), 45,
                                                zorder=10,
                                                color=self.M_PINK_H,
                                                marker=self.markerAltAz(),
                                                markersize=25,
                                                linestyle='none',
                                                fillstyle='none',
                                                clip_on=False,
                                                visible=False,
                                                )

        else:
            self.pointerAltAz, = axes.plot(180, 45,
                                           zorder=10,
                                           color=self.M_PINK_H,
                                           marker=self.markerAltAz(),
                                           markersize=25,
                                           linestyle='none',
                                           fillstyle='none',
                                           clip_on=False,
                                           visible=False,
                                           )

            self.pointerDome = mpatches.Rectangle((165, 1),
                                                  30,
                                                  88,
                                                  zorder=0,
                                                  color=self.M_GREY_MID,
                                                  lw=3,
                                                  clip_on=True,
                                                  fill=True,
                                                  visible=False)
            axes.add_patch(self.pointerDome)

        return True

    def drawAlignmentStars(self, axes=None):
        """
        drawAlignmentStars is rendering the alignment star map. this moves over time with
        the speed of earth turning. so we have to update the rendering, but on low speed
        without having any user interaction.

        :param axes: matplotlib axes object
        :return: true for test purpose
        """

        self.starsAlignAnnotate = list()
        hip = self.app.hipparcos
        hip.calculateAlignStarPositionsAltAz()

        self.starsAlign, = axes.plot(hip.az,
                                     hip.alt,
                                     marker=self.markerStar(),
                                     markersize=7,
                                     linestyle='None',
                                     color=self.MODE[self.operationMode]['starColor'],
                                     zorder=30,
                                     )

        for alt, az, name in zip(hip.alt, hip.az, hip.name):
            annotation = axes.annotate(name,
                                       xy=(az, alt),
                                       xytext=(2, 2),
                                       textcoords='offset points',
                                       xycoords='data',
                                       color=self.M_WHITE_L,
                                       zorder=30,
                                       fontsize=12,
                                       clip_on=True,
                                       )
            self.starsAlignAnnotate.append(annotation)

        return True

    def drawHemisphere(self):
        """
        drawHemisphere is the basic renderer for all items and widgets in the hemisphere
        window. it takes care of drawing the grid, enables three layers of transparent
        widgets for static content, moving content and star maps. this is mainly done to
        get a reasonable performance when redrawing the canvas. in addition it initializes
        the objects for points markers, patches, lines etc. for making the window nice
        and user friendly.
        the user interaction on the hemisphere windows is done by the event handler of
        matplotlib itself implementing an on Mouse handler, which takes care of functions.

        :return: True for test purpose
        """
        if self.closingWindow:
            return False

        axe, _ = self.generateFlat(widget=self.hemisphereMat,
                                   horizon=True)
        axeMove, _ = self.generateFlat(widget=self.hemisphereMatMove,
                                       horizon=True,
                                       showAxes=False)
        hasPolar = self.ui.showPolar.isChecked()
        if hasPolar:
            axePolar, _ = self.generatePolar(widget=self.polarMat,
                                             horizon=True)
            axePolarMove, _ = self.generatePolar(widget=self.polarMatMove,
                                                 horizon=True,
                                                 showAxes=False)
        else:
            self.pointerPolarAltAz = None
            self.pointsBuildAnnotate = None
            self.pointsPolarBuild = None
            self.pointsPolarBuildAnnotate = None
            self.polarMat.figure.clf()
            self.polarMatMove.figure.clf()

        axe.figure.canvas.flush_events()
        g = self.ui.hemisphere.geometry()
        self.ui.hemisphereMove.setGeometry(QRect(0, 0, g.width(), g.height()))

        if hasPolar:
            axePolar.figure.canvas.flush_events()
            g = self.ui.polar.geometry()
            self.ui.polarMove.setGeometry(QRect(0, 0, g.width(), g.height()))

        if self.ui.checkShowAlignStar.isChecked():
            self.drawAlignmentStars(axes=axe)

        else:
            self.starsAlign = None
            self.starsAlignAnnotate = None

        self.drawHemisphereStatic(axes=axe)
        self.drawHemisphereMoving(axes=axeMove)
        axe.figure.canvas.draw()

        if hasPolar:
            self.drawHemisphereStatic(axes=axePolar, polar=True)
            self.drawHemisphereMoving(axes=axePolarMove, polar=True)
            axePolar.figure.canvas.draw()

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries
import zlib
import os

# external packages
import astropy.io.fits as fits

# local imports
from base.indiClass import IndiClass


class CameraIndi(IndiClass):
    """
    the class Camera inherits all information and handling of the Camera device.


        >>> c = CameraIndi(app=None, signals=None, data=None)
    """

    __all__ = ['CameraIndi',
               ]

    # update rate to 1000 milli seconds for setting indi server
    UPDATE_RATE = 1000

    def __init__(self, app=None, signals=None, data=None):
        super().__init__(app=app, data=data, threadPool=app.threadPool)

        self.signals = signals
        self.data = data
        self.imagePath = ''
        self.isDownloading = False

    def setUpdateConfig(self, deviceName):
        """
        _setUpdateRate corrects the update rate of camera devices to get an defined
        setting regardless, what is setup in server side.

        :param deviceName:
        :return: success
        """

        if deviceName != self.deviceName:
            return False

        if self.device is None:
            return False

        # set BLOB mode also
        self.client.setBlobMode(blobHandling='Also',
                                deviceName=deviceName)

        # setting a object name
        objectName = self.device.getText('FITS_HEADER')
        objectName['FITS_OBJECT'] = 'skymodel'
        self.client.sendNewText(deviceName=deviceName,
                                propertyName='FITS_HEADER',
                                elements=objectName,
                                )

        # setting WCS Control off
        wcs = self.device.getSwitch('WCS_CONTROL')
        wcs['WCS_DISABLE'] = 'On'

        self.client.sendNewSwitch(deviceName=deviceName,
                                  propertyName='WCS_CONTROL',
                                  elements=wcs,
                                  )

        # setting active device for telescope
        telescope = self.device.getText('ACTIVE_DEVICES')
        telescope['ACTIVE_TELESCOPE'] = 'LX200 10micron'
        self.client.sendNewText(deviceName=deviceName,
                                propertyName='ACTIVE_DEVICES',
                                elements=telescope,
                                )

        # setting polling updates in driver
        update = self.device.getNumber('POLLING_PERIOD')

        if 'PERIOD_MS' not in update:
            return False
        if update.get('PERIOD_MS', 0) == self.UPDATE_RATE:
            return True

        update['PERIOD_MS'] = self.UPDATE_RATE
        suc = self.client.sendNewNumber(deviceName=deviceName,
                                        propertyName='POLLING_PERIOD',
                                        elements=update,
                                        )

        return suc

    def setExposureState(self):
        """
        :return: success
        """
        if not hasattr(self.device, 'CCD_EXPOSURE'):
            return False

        value = self.data.get('CCD_EXPOSURE.CCD_EXPOSURE_VALUE', 0)

        if self.device.CCD_EXPOSURE['state'] == 'Idle':
            self.signals.message.emit('')

        elif self.device.CCD_EXPOSURE['state'] == 'Busy':
            if value == 0:
                self.isDownloading = True
                self.signals.message.emit('download')
                self.signals.integrated.emit()

            else:
                self.signals.message.emit(f'expose {value:2.0f} s')

        elif self.device.CCD_EXPOSURE['state'] == 'Ok':
            self.signals.message.emit('')

        if self.device.CCD_EXPOSURE['state'] in ['Idle', 'Ok'] and self.isDownloading:
            self.isDownloading = False

        return True

    def updateNumber(self, deviceName, propertyName):
        """
        updateNumber is called whenever a new number is received in client. it runs
        through the device list and writes the number data to the according locations.

        :param deviceName:
        :param propertyName:
        :return:
        """
        if not super().updateNumber(deviceName, propertyName):
            return False

        if propertyName == 'CCD_EXPOSURE':
            self.setExposureState()

        return True

    def updateBLOB(self, deviceName, propertyName):
        """
        updateBLOB is called whenever a new BLOB is received in client. it runs
        through the device list and writes the number data to the according locations.

        :param deviceName:
        :param propertyName:
        :return: success
        """

        if not super().updateBLOB(deviceName, propertyName):
            return False

        data = self.device.getBlob(propertyName)

        if 'value' not in data:
            return False
        if 'name' not in data:
            return False
        if 'format' not in data:
            return False
        if data.get('name', '') != 'CCD1':
            return False
        if not self.imagePath:
            return False
        if not os.path.isdir(os.path.dirname(self.imagePath)):
            return False

        if data['format'] == '.fits.fz':
            HDU = fits.HDUList.fromstring(data['value'])
            fits.writeto(self.imagePath, HDU[0].data, HDU[0].header, overwrite=True)
            self.log.info('Image BLOB is in FPacked format')

        elif data['format'] == '.fits.z':
            HDU = fits.HDUList.fromstring(zlib.decompress(data['value']))
            fits.writeto(self.imagePath, HDU[0].data, HDU[0].header, overwrite=True)
            self.log.info('Image BLOB is compressed fits format')

        elif data['format'] == '.fits':
            HDU = fits.HDUList.fromstring(data['value'])
            fits.writeto(self.imagePath, HDU[0].data, HDU[0].header, overwrite=True)
            self.log.info('Image BLOB is uncompressed fits format')

        else:
            self.log.info('Image BLOB is not supported')

        self.signals.saved.emit(self.imagePath)
        return True

    def sendDownloadMode(self, fastReadout=False):
        """
        setDownloadMode sets the readout speed of the camera

        :return: success
        """

        if not self.device:
            return False

        # setting fast mode:
        quality = self.device.getSwitch('READOUT_QUALITY')
        self.log.debug(f'camera has readout quality entry: {quality}')

        if fastReadout:
            quality['QUALITY_LOW'] = 'On'
            quality['QUALITY_HIGH'] = 'Off'

        else:
            quality['QUALITY_LOW'] = 'Off'
            quality['QUALITY_HIGH'] = 'On'

        suc = self.client.sendNewSwitch(deviceName=self.deviceName,
                                        propertyName='READOUT_QUALITY',
                                        elements=quality,
                                        )

        return suc

    def expose(self,
               imagePath='',
               expTime=3,
               binning=1,
               fastReadout=True,
               posX=0,
               posY=0,
               width=1,
               height=1,
               focalLength=1,
               ):
        """

        :param imagePath:
        :param expTime:
        :param binning:
        :param fastReadout:
        :param posX:
        :param posY:
        :param width:
        :param height:
        :param focalLength:
        :return: success
        """

        if not self.device:
            return False

        self.imagePath = imagePath

        suc = self.sendDownloadMode(fastReadout=fastReadout)
        if not suc:
            self.log.debug('Download quality could not be set')

        indiCmd = self.device.getNumber('CCD_BINNING')
        indiCmd['HOR_BIN'] = binning
        indiCmd['VER_BIN'] = binning
        suc = self.client.sendNewNumber(deviceName=self.deviceName,
                                        propertyName='CCD_BINNING',
                                        elements=indiCmd,
                                        )
        if not suc:
            self.log.debug('Binning could not be set')

        indiCmd = self.device.getNumber('CCD_FRAME')
        indiCmd['X'] = posX
        indiCmd['Y'] = posY
        indiCmd['WIDTH'] = width
        indiCmd['HEIGHT'] = height
        suc = self.client.sendNewNumber(deviceName=self.deviceName,
                                        propertyName='CCD_FRAME',
                                        elements=indiCmd,
                                        )
        if not suc:
            self.log.debug('Frame could not be set')

        indiCmd = self.device.getNumber('CCD_EXPOSURE')
        indiCmd['CCD_EXPOSURE_VALUE'] = expTime
        suc = self.client.sendNewNumber(deviceName=self.deviceName,
                                        propertyName='CCD_EXPOSURE',
                                        elements=indiCmd,
                                        )
        return suc

    def abort(self):
        """
        abort cancels the exposing

        :return: success
        """

        if not self.device:
            return False

        indiCmd = self.device.getSwitch('CCD_ABORT_EXPOSURE')

        if 'ABORT' not in indiCmd:
            return False

        indiCmd['ABORT'] = 'On'

        suc = self.client.sendNewSwitch(deviceName=self.deviceName,
                                        propertyName='CCD_ABORT_EXPOSURE',
                                        elements=indiCmd,
                                        )

        return suc

    def sendCoolerSwitch(self, coolerOn=False):
        """
        sendCoolerTemp send the desired cooler temp, but does not switch on / off the cooler

        :param coolerOn:
        :return: success
        """

        if not self.device:
            return False

        # setting fast mode:
        cooler = self.device.getSwitch('CCD_COOLER')

        if coolerOn:
            cooler['COOLER_ON'] = 'On'
            cooler['COOLER_OFF'] = 'Off'

        else:
            cooler['COOLER_ON'] = 'Off'
            cooler['COOLER_OFF'] = 'On'

        suc = self.client.sendNewSwitch(deviceName=self.deviceName,
                                        propertyName='CCD_COOLER',
                                        elements=cooler,
                                        )

        return suc

    def sendCoolerTemp(self, temperature=0):
        """
        sendCoolerTemp send the desired cooler temp, indi does automatically start cooler

        :param temperature:
        :return: success
        """

        if not self.device:
            return False

        temp = self.device.getNumber('CCD_TEMPERATURE')

        if 'CCD_TEMPERATURE_VALUE' not in temp:
            return False

        temp['CCD_TEMPERATURE_VALUE'] = temperature
        suc = self.client.sendNewNumber(deviceName=self.deviceName,
                                        propertyName='CCD_TEMPERATURE',
                                        elements=temp,
                                        )

        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries
import os
import logging
import platform

# external packages
try:
    from pywinauto import timings
except Exception:
    hasAutomation = False
else:
    hasAutomation = True

from PyQt5.QtCore import QObject
from pywinauto.findwindows import find_windows
from pywinauto.application import AppStartError, Application
import pywinauto.controls.win32_controls as controls
import winreg
from winreg import HKEY_LOCAL_MACHINE

# local imports


class AutomateWindows(QObject):
    __all__ = ['AutomateWindows',
               ]

    log = logging.getLogger(__name__)

    UPDATER_EXE = 'GmQCIv2.exe'
    UTC_1_FILE = 'finals.data'
    UTC_2_FILE = 'tai-utc.dat'

    COMET_FIELDS = [
        'Orbit_type',
        'Provisional_packed_desig',
        'Year_of_perihelion',
        'Month_of_perihelion',
        'Day_of_perihelion',
        'Perihelion_dist',
        'e',
        'Peri',
        'Node',
        'i',
        'Epoch_year',
        'Epoch_month',
        'Epoch_day',
        'H',
        'G',
        'Designation_and_name',
        'Ref'
    ]

    ASTEROID_FIELDS = [
        'H',
        'G',
        'Num_obs',
        'rms',
        'U',
        'Arc_years',
        'Perturbers',
        'Perturbers_2',
        'Principal_desig',
        'Epoch',
        'M',
        'Peri',
        'Node',
        'i',
        'e',
        'n',
        'a',
        'Ref',
        'Num_opps',
        'Computer',
        'Hex_flags',
        'Last_obs',
        'Tp',
        'Orbital_period',
        'Perihelion_dist',
        'Aphelion_dist',
        'Semilatus_rectum',
        'Synodic_period',
        'Orbit_type'
    ]

    def __init__(self, app):
        super().__init__()
        self.app = app
        self.threadPool = app.threadPool

        if not hasAutomation:
            self.installPath = ''
            self.name = ''
            self.available = False
            return

        val = self.getAppSettings(['10micron QCI control',
                                   '10micron control'])
        self.log.debug(f'QCI Updater settings: [{val}]')
        self.available, self.name, self.installPath = val
        self.updaterRunnable = self.installPath + self.UPDATER_EXE
        self.updater = None
        self.actualWorkDir = os.getcwd()

    @staticmethod
    def getRegistryPath():
        """
        :return:
        """
        if platform.machine().endswith('64'):
            regPath = 'SOFTWARE\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\Uninstall'

        else:
            regPath = 'SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall'

        return regPath

    @staticmethod
    def convertRegistryEntryToDict(subkey):
        """
        :param subkey:
        :return:
        """
        values = dict()

        for j in range(0, winreg.QueryInfoKey(subkey)[1]):
            values[winreg.EnumValue(subkey, j)[0]] = winreg.EnumValue(subkey, j)[1]

        return values

    def getValuesForNameKeyFromRegistry(self, nameKey):
        """
        :param nameKey:
        :return:
        """
        regPath = self.getRegistryPath()
        key = winreg.OpenKey(HKEY_LOCAL_MACHINE, regPath)
        subkey = winreg.OpenKey(key, nameKey)
        values = self.convertRegistryEntryToDict(subkey)
        winreg.CloseKey(subkey)
        winreg.CloseKey(key)

        return values

    @staticmethod
    def searchNameInRegistry(appName, key):
        """
        :param appName:
        :param key:
        :return:
        """
        for i in range(0, winreg.QueryInfoKey(key)[0]):
            nameKey = winreg.EnumKey(key, i)
            if appName in nameKey:
                break
        else:
            nameKey = ''

        return nameKey

    def getNameKeyFromRegistry(self, appName):
        """
        :param appName:
        :return:
        """
        regPath = self.getRegistryPath()
        key = winreg.OpenKey(HKEY_LOCAL_MACHINE, regPath)
        nameKey = self.searchNameInRegistry(appName, key)
        winreg.CloseKey(key)

        return nameKey

    def extractPropertiesFromRegistry(self, appName):
        """
        :param appName:
        :return:
        """
        nameKey = self.getNameKeyFromRegistry(appName)
        if not appName:
            return False, '', ''

        values = self.getValuesForNameKeyFromRegistry(nameKey)
        if 'InstallLocation' not in values:
            self.log.warning('QCI updater not found.')
            return False, '', ''

        if appName in values.get('DisplayName', ''):
            available = True
            name = values['DisplayName']
            installPath = values['InstallLocation']

        else:
            available = False
            installPath = ''
            name = ''
            self.log.warning('QCI updater not found.')

        return available, installPath, name

    def cycleThroughAppNames(self, appNames):
        """
        :param appNames:
        :return:
        """
        for appName in appNames:
            val = self.extractPropertiesFromRegistry(appName)
            if val[0]:
                break
        else:
            return False, '', ''

        return val

    def getAppSettings(self, appNames):
        """
        :param appNames:
        :return:
        """
        try:
            val = self.cycleThroughAppNames(appNames)
            available, installPath, displayName = val

        except Exception as e:
            self.log.debug(f'{e}')
            return False, '', ''

        return available, displayName, installPath

    def checkFloatingPointErrorWindow(self):
        """
        :return:
        """
        try:
            dialog = timings.wait_until_passes(2,
                                               0.2,
                                               lambda: find_windows(title='GmQCIv2',
                                                                    class_name='#32770')[0])
            winOK = self.updater.window(handle=dialog)
            winOK['OK'].click()

        except timings.TimeoutError:
            return True

        except Exception as e:
            self.log.error(f'error{e}')
            return False

        else:
            return True

    def startUpdater(self):
        """
        :return:
        """
        if platform.architecture()[0] == '32bit':
            self.updater = Application(backend='win32')
            self.log.info('Using 32Bit backend win32')

        else:
            self.updater = Application(backend='uia')
            self.log.info('Using 64Bit backend uia')

        try:
            self.updater.start(self.installPath + self.UPDATER_EXE)

        except AppStartError:
            self.log.error('Failed to start updater, please check!')
            return False

        except Exception as e:
            self.log.error(f'Failed to start updater, error {e}')
            return False

        else:
            suc = self.checkFloatingPointErrorWindow()
            return suc

    def clearUploadMenuCommands(self):
        """
        :return:
        """
        win = self.updater['10 micron control box update']
        win['next'].click()
        win['next'].click()
        controls.ButtonWrapper(win['Control box firmware']).uncheck_by_click()
        controls.ButtonWrapper(win['Orbital parameters of comets']).uncheck_by_click()
        controls.ButtonWrapper(win['Orbital parameters of asteroids']).uncheck_by_click()
        controls.ButtonWrapper(win['Orbital parameters of satellites']).uncheck_by_click()
        controls.ButtonWrapper(win['UTC / Earth rotation data']).uncheck_by_click()
        return True

    def clearUploadMenu(self):
        """
        :return:
        """
        try:
            self.clearUploadMenuCommands()

        except Exception as e:
            self.log.error('error{0}'.format(e))
            return False

        return True

    def prepareUpdater(self):
        """
        :return:
        """
        if not self.installPath:
            self.log.error(f'No updater found: {self.installPath}')
            return False

        self.updater = None
        os.chdir(os.path.dirname(self.installPath))

        suc = self.startUpdater()
        if not suc:
            os.chdir(self.actualWorkDir)
            return False

        suc = self.clearUploadMenu()
        if not suc:
            os.chdir(self.actualWorkDir)
            return False

        return True

    def doUploadAndCloseInstallerCommands(self):
        """
        :return:
        """
        win = self.updater['10 micron control box update']
        win['next'].click()
        win['next'].click()
        win['Update Now'].click()
        return True

    def pressOK(self):
        """
        :return:
        """
        dialog = timings.wait_until_passes(60,
                                           0.5,
                                           lambda: find_windows(title='Update completed',
                                                                class_name='#32770')[0])
        winOK = self.updater.window(handle=dialog)
        winOK['OK'].click()

        return True

    def doUploadAndCloseInstaller(self):
        """
        :return:
        """
        try:
            self.doUploadAndCloseInstallerCommands()
            self.pressOK()

        except Exception as e:
            self.log.error('error{0}'.format(e))
            return False

        return True

    def uploadMPCDataCommands(self, comets=False):
        """
        :param comets:
        :return:
        """
        win = self.updater['10 micron control box update']
        if comets:
            controls.ButtonWrapper(win['Orbital parameters of comets']).check_by_click()
            win['Edit...4'].click()
            popup = self.updater['Comet orbits']

        else:
            controls.ButtonWrapper(win['Orbital parameters of asteroids']).check_by_click()
            win['Edit...3'].click()
            popup = self.updater['Asteroid orbits']

        popup['MPC file'].click()
        filedialog = self.updater['Dialog']
        text = self.installPath + 'minorPlanets.mpc'
        controls.EditWrapper(filedialog['File &name:Edit']).set_edit_text(text)
        if platform.architecture()[0] == '32bit':
            filedialog['Button16'].click()

        else:
            filedialog['OpenButton4'].click()

        popup['Close'].click()
        return True

    def uploadMPCData(self, comets=False):
        """
        :param comets:
        :return:
        """
        try:
            self.prepareUpdater()
            self.uploadMPCDataCommands(comets=comets)

        except Exception as e:
            self.log.error(f'error{e}')
            return False

        else:
            suc = self.doUploadAndCloseInstaller()
            return suc

        finally:
            os.chdir(self.actualWorkDir)

    def uploadEarthRotationDataCommands(self):
        """
        :return:
        """
        win = self.updater['10 micron control box update']
        controls.ButtonWrapper(win['UTC / Earth rotation data']).check_by_click()
        win['Edit...1'].click()
        popup = self.updater['UTC / Earth rotation data']
        popup['Import files...'].click()
        filedialog = self.updater['Open finals data']
        text = self.installPath + self.UTC_1_FILE
        controls.EditWrapper(filedialog['File &name:Edit']).set_text(text)
        if platform.architecture()[0] == '32bit':
            filedialog['Button16'].click()
        else:
            filedialog['OpenButton4'].click()
        filedialog = self.updater['Open tai-utc.dat']
        text = self.installPath + self.UTC_2_FILE
        controls.EditWrapper(filedialog['File &name:Edit']).set_text(text)
        if platform.architecture()[0] == '32bit':
            filedialog['Button16'].click()

        else:
            filedialog['OpenButton4'].click()

        fileOK = self.updater['UTC data']
        fileOK['OK'].click()
        return True

    def uploadEarthRotationData(self):
        """
        :return:
        """
        self.prepareUpdater()

        try:
            self.uploadEarthRotationDataCommands()

        except Exception as e:
            self.log.error(f'error{e}')
            os.chdir(self.actualWorkDir)
            return False

        else:
            suc = self.doUploadAndCloseInstaller()
            return suc

        finally:
            os.chdir(self.actualWorkDir)

    def uploadTLEDataCommands(self):
        """
        :return:
        """
        win = self.updater['10 micron control box update']

        controls.ButtonWrapper(win['Orbital parameters of satellites']).check_by_click()
        win['Edit...2'].click()
        popup = self.updater['Satellites orbits']
        popup['Load from file'].click()
        filedialog = self.updater['Dialog']
        text = self.installPath + 'satellites.tle'
        controls.EditWrapper(filedialog['File &name:Edit']).set_text(text)
        if platform.architecture()[0] == '32bit':
            filedialog['Button16'].click()

        else:
            filedialog['OpenButton4'].click()

        popup['Close'].click()

        return True

    def uploadTLEData(self):
        """
        :return:
        """
        self.prepareUpdater()

        try:
            self.uploadTLEDataCommands()

        except Exception as e:
            self.log.error(f'error{e}')
            os.chdir(self.actualWorkDir)
            return False

        else:
            suc = self.doUploadAndCloseInstaller()
            return suc

        finally:
            os.chdir(self.actualWorkDir)



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries
import platform
import os
import re

# external packages
from PyQt5.QtWidgets import QWidget, QDesktopWidget, QFileDialog, QMessageBox
from PyQt5.QtGui import QPalette, QIcon, QPixmap
from PyQt5.QtCore import QSortFilterProxyModel, QDir, QObject, pyqtSignal
from PyQt5.QtCore import Qt, QSize, QEvent
from skyfield.api import Angle

# local imports
from gui.utilities.stylesQtCss import Styles
from gui.utilities.toolsMatplotlib import ToolsMatplotlib
from mountcontrol.convert import sexagesimalizeToInt

__all__ = [
    'MWidget',
    'FileSortProxyModel',
    'QMultiWait',
]


class FileSortProxyModel(QSortFilterProxyModel):
    """
    FileSortProxyModel enables a proxy solution for reversing the order of all
    file dialogues. The sorting is now Descending meaning the last added files
    will be on top. This is don by just overwriting the sort method
    """

    def sort(self, column, order):
        self.sourceModel().sort(0, Qt.DescendingOrder)


class QMultiWait(QObject):
    """
    QMultiWaitable implements a signal collection class for waiting of entering
    multiple signals before firing the "AND" relation of all signals.
    derived from:

    https://stackoverflow.com/questions/21108407/qt-how-to-wait-for-multiple-signals

    in addition all received signals could be reset
    """

    ready = pyqtSignal()

    def __init__(self):
        super().__init__()
        self.waitable = set()
        self.waitready = set()

    def addWaitableSignal(self, signal):
        if signal not in self.waitable:
            self.waitable.add(signal)
            signal.connect(self.checkSignal)

    def checkSignal(self):
        sender = self.sender()
        self.waitready.add(sender)

        if len(self.waitready) == len(self.waitable):
            self.ready.emit()

    def resetSignals(self):
        self.waitready = set()

    def clear(self):
        for signal in self.waitable:
            signal.disconnect(self.checkSignal)

        self.waitable = set()
        self.waitready = set()


class MWidget(QWidget, Styles, ToolsMatplotlib):
    """
    MWidget defines the common parts for all windows used in MountWizzard 4 and
    extends the standard widgets. All widgets configs which are used mor than
    one time are centralized in this class.

    For the File dialogues, the original widgets are used, but with the removal
    of some features to make them simpler. As one optimization they always show
    the files and directories in descending order.

    The styles of the widgets are defined separately in a css looking
    stylesheet. The standard screen size will be 800x600 pixel for all windows,
    but except for the main one are sizable.
    """

    __all__ = ['MWidget',
               ]

    def __init__(self):
        super().__init__()

        self.palette = QPalette()
        self.initUI()
        self.screenSizeX = QDesktopWidget().screenGeometry().width()
        self.screenSizeY = QDesktopWidget().screenGeometry().height()
        self.setAttribute(Qt.WA_DeleteOnClose, True)

        newFlag = Qt.CustomizeWindowHint | Qt.WindowSystemMenuHint
        newFlag = newFlag | Qt.WindowMinimizeButtonHint | Qt.WindowMaximizeButtonHint
        self.setWindowFlags(self.windowFlags() | newFlag)

    @staticmethod
    def wIcon(gui=None, name=''):
        """
        widget icon adds an icon to a gui element like an button.

        :param      gui:        gui element, which will be expanded by an icon
        :param      name:       icon to be added
        :return:    true for test purpose
        """
        if not gui:
            return False
        if not name:
            return False

        icon = QIcon(f':/icon/{name}.svg')
        gui.setIcon(icon)
        gui.setIconSize(QSize(16, 16))
        gui.setProperty('alignLeft', True)
        gui.style().unpolish(gui)
        gui.style().polish(gui)
        return True

    def getStyle(self):
        """
        getStyle return the actual stylesheet for the used platform.
        As the font sizes vary between Darwin and Windows / Ubuntu there were
        two sets of stylesheets used. a basic stylesheet adds undifferentiated
        properties.

        :return:    actual stylesheet string
        """
        if platform.system() == 'Darwin':
            return self.MAC_STYLE + self.BASIC_STYLE

        else:
            return self.NON_MAC_STYLE + self.BASIC_STYLE

    def initUI(self):
        """
        init_UI makes the basic initialisation of the GUI. is sets the window
        flags and sets the handling of the window. is as well fixes the windows
        size (unless a windows will be scalable later on. in addition the
        appropriate style sheet for mac and non mac will be selected and used.

        :return:    true for test purpose
        """
        self.setWindowFlags(self.windowFlags())
        style = self.getStyle()
        self.setStyleSheet(style)
        self.setMouseTracking(True)
        self.setWindowIcon(QIcon(':/mw4.ico'))
        return True

    @staticmethod
    def changeStyleDynamic(widget=None, widgetProperty=None, value=None):
        """
        changeStyleDynamic changes the stylesheet of a given uii element and
        makes it visible. therefore the element has to be unpolished and
        polished again.

        :param      widget: widget element, where the stylesheet has to
                            be changed
        :param      widgetProperty: stylesheet attribute which has to be
                                    changes
        :param      value:  new value of the attribute
        :return:    true for test purpose
        """
        if not widget:
            return False
        if not widgetProperty:
            return False
        if value is None:
            return False
        if widget.property(widgetProperty) == value:
            return True

        widget.style().unpolish(widget)
        widget.setProperty(widgetProperty, value)
        widget.style().polish(widget)

        return True

    @staticmethod
    def extractNames(names=''):
        """
        extractNames splits a given path to basename and extension
        if the input is a multiple list, there will be as return values a
        multiple list as well, otherwise single values

        :param      names:   full path of file (s)
        :return:    short:  basename without extension
                    ext:    extension
                    name:   name
        """
        if not names:
            return '', '', ''
        if not isinstance(names, list):
            return '', '', ''

        shortList = list()
        extList = list()
        nameList = list()

        for name in names:
            if name:
                short, ext = os.path.splitext(name)
                short = os.path.basename(short)

            else:
                short = ''
                ext = ''

            nameList.append(os.path.abspath(name))
            shortList.append(short)
            extList.append(ext)

        if len(names) == 1:
            return nameList[0], shortList[0], extList[0]

        else:
            return nameList, shortList, extList

    def prepareFileDialog(self, window=None, enableDir=False, reverseOrder=False):
        """
        prepareFileDialog does some tweaking of the standard file dialogue
        widget for geometry and general settings. it also removes some parts and
        makes the dialog modal.

        :param window:  parent class
        :param enableDir:   allows dir selection in file box
        :param reverseOrder:   file selection z->a
        :return:        dlg, the dialog widget
        """
        if not window:
            return None

        dlg = QFileDialog()
        dlg.setOptions(QFileDialog.DontUseNativeDialog)
        dlg.setWindowIcon(QIcon(':/mw4.ico'))
        dlg.setStyleSheet(self.getStyle())
        dlg.setViewMode(QFileDialog.List)
        dlg.setModal(True)
        if reverseOrder:
            dlg.setProxyModel(FileSortProxyModel(self))

        if enableDir:
            dlg.setFilter(QDir.Files | QDir.AllDirs)

        else:
            dlg.setFilter(QDir.Files)

        width = 600
        height = 400

        ph = window.geometry().height()
        pw = window.geometry().width()
        px = window.geometry().x()
        py = window.geometry().y()

        dlg.setGeometry(px + int(0.5 * (pw - width)),
                        py + int(0.5 * (ph - height)),
                        width,
                        height)

        return dlg

    @staticmethod
    def runDialog(dlg):
        """
        :param dlg:
        :return: result
        """
        return dlg.exec_()

    def messageDialog(self, parentWidget, title, question):
        """
        :param parentWidget:
        :param title:
        :param question:
        :return: OK
        """
        msg = QMessageBox()
        msg.setWindowModality(Qt.ApplicationModal)
        msg.setStyleSheet(self.getStyle())
        msg.setWindowTitle(title)
        pixmap = QPixmap(':/icon/question.svg').scaled(64, 64)
        msg.setIconPixmap(pixmap)
        msg.setText(question)
        msg.setStandardButtons(msg.Yes | msg.No)
        msg.setDefaultButton(msg.No)
        msg.show()
        x = parentWidget.x() + int((parentWidget.width() - msg.width()) / 2)
        y = parentWidget.y() + int((parentWidget.height() - msg.height()) / 2)
        msg.move(x, y)
        reply = self.runDialog(msg)

        if reply != msg.Yes:
            return False

        else:
            return True

    def openFile(self,
                 window=None,
                 title='',
                 folder='',
                 filterSet=None,
                 enableDir=False,
                 multiple=False,
                 reverseOrder=False):
        """
        openFile handles a single file select with filter in a non native format.

        :param window:      parent window class
        :param title:       title for the file dialog
        :param folder:      starting folder for searching the file
        :param filterSet:   file extension filter
        :param enableDir:   allows dir selection in file box
        :param multiple :   allows multiple selection in file box
        :param reverseOrder :   file selection z->a
        :return:            name: full path for file else empty
                            short: just file name without extension
                            ext: extension of the file
        """

        if not window:
            return '', '', ''

        if not title:
            return '', '', ''

        if not folder:
            return '', '', ''

        if not filterSet:
            return '', '', ''

        dlg = self.prepareFileDialog(window=window,
                                     enableDir=enableDir,
                                     reverseOrder=reverseOrder)
        dlg.setAcceptMode(QFileDialog.AcceptOpen)

        dlg.setWindowTitle(title)
        dlg.setNameFilter(filterSet)
        dlg.setDirectory(folder)

        if multiple:
            dlg.setFileMode(QFileDialog.ExistingFiles)

        else:
            dlg.setFileMode(QFileDialog.ExistingFile)

        result = self.runDialog(dlg)

        if not result:
            return '', '', ''

        filePath = dlg.selectedFiles()
        full, short, ext = self.extractNames(names=filePath)

        return full, short, ext

    def saveFile(self,
                 window=None,
                 title='',
                 folder='',
                 filterSet=None,
                 enableDir=False):
        """
        saveFile handles a single file save with filter in a non native format.

        :param window:      parent window class
        :param title:       title for the file dialog
        :param folder:      starting folder for searching the file
        :param filterSet:   file extension filter
        :param enableDir:   allows dir selection in file box
        :return:            name: full path for file else empty
                            short: just file name without extension
                            ext: extension of the file
        """

        if not window:
            return '', '', ''

        if not title:
            return '', '', ''

        if not folder:
            return '', '', ''

        if not filterSet:
            return '', '', ''

        dlg = self.prepareFileDialog(window, enableDir)
        dlg.setAcceptMode(QFileDialog.AcceptSave)

        dlg.setWindowTitle(title)
        dlg.setNameFilter(filterSet)
        dlg.setDirectory(folder)

        result = self.runDialog(dlg)

        if not result:
            return '', '', ''

        filePath = dlg.selectedFiles()
        full, short, ext = self.extractNames(names=filePath)

        return full, short, ext

    def openDir(self,
                window=None,
                title='',
                folder=''):
        """
        openFile handles a single file select with filter in a non native format.

        :param window:      parent window class
        :param title:       title for the file dialog
        :param folder:      starting folder for searching the file
        :return:            name: full path for file else empty
                            short: just file name without extension
                            ext: extension of the file
        """

        if not window:
            return '', '', ''

        if not title:
            return '', '', ''

        if not folder:
            return '', '', ''

        dlg = self.prepareFileDialog(window=window, enableDir=True)
        dlg.setAcceptMode(QFileDialog.AcceptOpen)

        dlg.setWindowTitle(title)
        dlg.setDirectory(folder)
        dlg.setFileMode(QFileDialog.DirectoryOnly)

        result = self.runDialog(dlg)

        if not result:
            return '', '', ''

        filePath = dlg.selectedFiles()
        full, short, ext = self.extractNames(names=filePath)

        return full, short, ext

    @staticmethod
    def clickable(widget=None):
        """
        It uses one filter object per label, which is created when the clickable()
        function is called with the widget that is to be click-enabled. The function
        returns a clicked() signal that actually belongs to the filter object. The caller
        can connect this signal to a suitable callable object.

        :param widget:      widget for what the event filter works
        :return:            filtered event
        """
        if not widget:
            return None

        class MouseDoubleClickEventFilter(QObject):
            clicked = pyqtSignal(object)

            def eventFilter(self, obj, event):
                if obj != widget:
                    return False

                if event.type() not in [QEvent.MouseButtonRelease,
                                        QEvent.FocusIn]:
                    return False

                if event.type() == QEvent.MouseButtonRelease:
                    if obj.rect().contains(event.pos()):
                        self.clicked.emit(widget)
                        return True

                return False

        doubleClickEventFilter = MouseDoubleClickEventFilter(widget)
        widget.installEventFilter(doubleClickEventFilter)

        return doubleClickEventFilter.clicked

    @staticmethod
    def guiSetText(ui, formatElement, value=None):
        """

        :param ui:
        :param formatElement:
        :param value:
        :return: True for test purpose
        """

        if not ui:
            return False

        if not formatElement:
            return False

        if value is None:
            text = '-'

        else:
            formatStr = '{0:' + formatElement + '}'
            text = formatStr.format(value)

        ui.setText(text)

        return True

    @staticmethod
    def returnDriver(sender, searchDict, addKey=''):
        """
        returnDriver takes the sender widget from a gui interaction and compares is to the
        widget objects of a search dicts to retrieve is original name.
        therefore we need to swap key value pais in the search dict as we make a reverse
        search.
        in addition to make it more usable the search dict might have some sub dicts where
        to find the gui elements. if given, the will be extracted on the forehand.

        :param sender:
        :param searchDict:
        :param addKey:
        :return:
        """

        if addKey:
            searchD = dict([(key, value[addKey]) for key, value in searchDict.items()])

        else:
            searchD = searchDict

        searchD = dict([(value, key) for key, value in searchD.items()])
        driver = searchD.get(sender, '')

        return driver

    @staticmethod
    def formatLatLonToAngle(value, pf):
        """
        :param value:
        :param pf: double character, first indicates the negative sign
        :return:
        """
        if value is None:
            return None

        value = value.strip()
        p1 = re.compile(r'(\d{1,3})([' + pf + r'])\s*(\d\d)?\s*(\d\d)?[.,]?(\d*)?')
        p2 = re.compile(r'([-+]?)(\d{1,3})[.,]?(\d*)?')
        isSexagesimal = p1.fullmatch(value) is not None
        isFloat = p2.fullmatch(value) is not None

        elements = p2.split(value)
        if isFloat:
            angle = float(value.replace(',', '.'))

        elif isSexagesimal:
            angle = float(elements[2])
            if len(elements) > 5:
                angle += float(elements[6]) / 60
            if len(elements) > 9:
                angle += float(elements[10]) / 3600
            if elements[4].startswith(pf[0]):
                angle = -angle

        else:
            return None

        if 'N' in pf:
            maxAbs = 90
        else:
            maxAbs = 180

        if angle > maxAbs:
            return None

        elif angle < -maxAbs:
            return None

        angle = Angle(degrees=angle)
        return angle

    def convertLatToAngle(self, value):
        """
        :param value:
        :return:
        """
        angle = self.formatLatLonToAngle(value, 'SN')
        return angle

    def convertLonToAngle(self, value):
        """
        :param value:
        :return:
        """
        angle = self.formatLatLonToAngle(value, 'WE')
        return angle

    @staticmethod
    def convertRaToAngle(value):
        """
        :param value:
        :return:
        """
        if value is None:
            return None

        value = value.strip()
        p1 = re.compile(r'([+-]?)(\d{1,3})H\s*(\d\d)?\s*(\d\d)?[.,]?(\d*)?')
        p2 = re.compile(r'([+-]?)(\d{1,3})\s+(\d\d)?\s*(\d\d)?[.,]?(\d*)?')
        p3 = re.compile(r'([+-]?)(\d{1,3})[.,]?(\d*)?')
        isP1 = p1.fullmatch(value) is not None
        isP2 = p2.fullmatch(value) is not None
        isSexagesimal = isP1 or isP2
        isFloat = p3.fullmatch(value) is not None

        if isP1:
            elements = p1.split(value)
        elif isP2:
            elements = p2.split(value)
        else:
            elements = ''

        if isFloat:
            angle = float(value.replace(',', '.'))

        elif isSexagesimal:
            angle = float(elements[2])
            if elements[3] is not None:
                angle += float(elements[3]) / 60
            if elements[4] is not None:
                angle += float(elements[4]) / 3600

        else:
            return None

        if angle >= 24:
            return None

        if angle < 0:
            return None

        if isFloat:
            angle = Angle(hours=angle / 360 * 24)

        elif isSexagesimal:
            angle = Angle(hours=angle)

        return angle

    @staticmethod
    def convertDecToAngle(value):
        """
        :param value:
        :return:
        """
        if value is None:
            return None

        value = value.strip()
        p1 = re.compile(r'([+-]?)(\d{1,3})Deg\s*(\d\d)?\s*(\d\d)?[.,]?(\d*)?')
        p2 = re.compile(r'([+-]?)(\d{1,3})\s+(\d\d)?\s*(\d\d)?[.,]?(\d*)?')
        p3 = re.compile(r'([+-]?)(\d{1,3})[.,]?(\d*)?')
        isP1 = p1.fullmatch(value) is not None
        isP2 = p2.fullmatch(value) is not None
        isSexagesimal = isP1 or isP2
        isFloat = p3.fullmatch(value) is not None

        if isP1:
            elements = p1.split(value)
        elif isP2:
            elements = p2.split(value)
        else:
            elements = ''

        if isFloat:
            angle = float(value.replace(',', '.'))

        elif isSexagesimal:
            angle = float(elements[2])
            if elements[3] is not None:
                angle += float(elements[3]) / 60
            if elements[4] is not None:
                angle += float(elements[4]) / 3600
            if elements[1].startswith('-'):
                angle = -angle
        else:
            return None

        if angle > 90:
            return None

        if angle < -90:
            return None

        angle = Angle(degrees=angle)
        return angle

    @staticmethod
    def formatHstrToText(angle):
        """
        :param angle:
        :return:
        """
        sgn, h, m, s, frac = sexagesimalizeToInt(angle.hours, 0)
        text = f'{h:02d} {m:02d} {s:02d}'
        return text

    @staticmethod
    def formatDstrToText(angle):
        """
        :param angle:
        :return:
        """
        sgn, h, m, s, frac = sexagesimalizeToInt(angle.degrees, 0)
        sign = '+' if sgn >= 0 else '-'
        text = f'{sign}{h:02d} {m:02d} {s:02d}'
        return text

    @staticmethod
    def formatLatToText(angle):
        """
        :param angle:
        :return:
        """
        sgn, h, m, s, frac = sexagesimalizeToInt(angle.degrees, 0)
        sign = 'N' if sgn >= 0 else 'S'
        text = f'{h:02d}{sign} {m:02d} {s:02d}'
        return text

    @staticmethod
    def formatLonToText(angle):
        """
        :param angle:
        :return:
        """
        sgn, h, m, s, frac = sexagesimalizeToInt(angle.degrees, 0)
        sign = 'E' if sgn >= 0 else 'W'
        text = f'{h:03d}{sign} {m:02d} {s:02d}'
        return text
